/*
 * Copyright (C) 1999 Dirk-Willem van Gulik, dirkx@webweaving.org 
 *	WebWeaving Consultancy, All rights reserved.
 *
 * Original version for FreeBSD; Changes for Linux by Brad Hards, 
 * brad.hards@dao.defence.gov.au. See '#ifdef BSD'.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY WEBWEAVING AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL WEBWEAVING OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 * 
 * This uploader should work with all with all USB-EZ Anchorchips which
 * have on board ram; i.e. the AN21y1, AN21y5 and AN21y6QC series. Where y
 * is the log2 of the internal memory size (i.e. aAN2146QC has 2^64=16kByte
 * of ram.)
 *
 * It has however only been tested; and written specifically for the
 * ActiveWire Inc. USB board. See 
 *	http://www.activewireinc.com
 * 
 * Anchorchips has excelent documentation on line at
 *     http://www.anchorchips.com
 */

#include <sys/param.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>

#include <sys/param.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#ifdef BSD
#include <dev/usb/usb.h>			/* FreeBSD */
#else
#include "/usr/src/linux/drivers/usb/usb.h"	/* Linux */
#endif

#define HEX 0
#define BIN 1
#define DOUTYPE HEX

void usage(char * s) {
	fprintf(stderr,"Syntax: %s [-v] [-a from [-e end]] "
			"[-s size] [-h|b] [-o outfile] device\n",s);
	exit(1);
}
extern char *optarg;
int 
main(int argc, char ** argv) 
{
	int fd; 
	int at = 0, en = 0;
	int s = 32 * 1024;
	FILE *fout = stdout;
#ifdef BSD
 	struct usb_ctl_request ur;
#else
	struct usb_proc_ctrltransfer ur;
#endif
	unsigned char crc, d[256];
     	int verbose=0, ch,outtype=DOUTYPE;

     	while ((ch = getopt(argc, argv, "hbva:e:s:o:")) != -1)
             switch (ch) {
	     case 'h':
		     outtype = HEX;
		     break;
	     case 'b':
		     outtype = BIN;
		     break;
             case 'v':
                     verbose ++;
                     break;
             case 'o':
                     if ((fout = fopen(optarg, "w")) < 0) {
                             fprintf(stderr,
                                 "%s: Could not open out file %s: %s\n", 
					argv[0], optarg, strerror(errno));
                             exit(1);
                     }
                     break;
	     case 'a':
	  	     at = strtol(optarg,NULL,0);
		     break;
	     case 'e':
	  	     en =  strtol(optarg,NULL,0);
		     break;
	     case 's':
		     s =  strtol(optarg,NULL,0);
		     break;
             case '?':
             default:
                     usage(argv[0]);
             }

	if (en && !at) {
		fprintf(stderr, "-e without -a\n");
		usage(argv[0]);
	}

	if (en)
		s = en - at;		/* calculate size */
	else
		en = at + s;		/* calculate end */

	if ((s < 0) || (at < 0) || (en < at))
		usage(argv[0]);

        argc -= optind;
        argv += optind;

	if (argc != 1)
		usage(argv[0]);

    	fd = open(argv[argc-1], O_WRONLY, 0);
        if ( fd == -1) {
                fprintf(stderr, "Open usb ctrl endpoint '%s' failed: %s\n",
                       	argv[1], strerror(errno));
                return(1);
        } else if (verbose)
		printf("; Control endpoint %s opened\n", argv[1]);

	/* Although USB does not limit you here; the Anchor docs
   	 * quote 64 as a limit; and Mato@activewireinc.com suggested
 	 * to use 16. 
 	 */
#define LEN 16


        if (verbose)
		printf("; Uploading from %04x to %04x (%d bytes) "
			"in chunks of %d bytes.\n",
			at, at+s, s,LEN);
	
	for(; at <= en; at += LEN) {
		int i, l=LEN;
		if (at+l > en) 
			l = en-at;

#ifdef BSD
		ur.request.bmRequestType= 0xC0;		/* see Anchor TRM v1.6 page 64 */
		ur.request.bRequest 	= 0xA0; 	/* Anchor Down/Upload */

		USETW(ur.request.wValue,at);		/* Starting Address */
		USETW(ur.request.wIndex,0);		/* unused */
		USETW(ur.request.wLength,l);		/* payload len in bytes */
	
		ur.flags = 0;
		ur.actlen = 0;
#else
                ur.requesttype= 0xC0;           /* see Anchor TRM v1.6 page 64 */
                ur.request      = 0xA0;         /* Anchor Down/Upload */
                ur.value = at;          /* Starting Address */
                ur.index = 0;           /* unused */
                ur.length = l;          /* payload len in bytes */
#endif
		ur.data	= (void *)d;

#ifdef BSD
        	if(ioctl(fd, USB_DO_REQUEST,&ur)) {
#else
		if(ioctl(fd, USB_PROC_CONTROL, &ur)) {
#endif
			fprintf(stderr,"Upload failed at adr %X: %s\nAborted\n",
				at,strerror(errno));
			break;
		}
#ifdef BSD
		if (ur.actlen != l) {
			fprintf(stderr,"Failed to get full %d chunk at adr %X: %s\nAborted\n",
				l,at,strerror(errno));
			break;
		}
#endif
		if(outtype == HEX) {
			/* and create an intel hex line...*/
			fprintf(fout,":%02X%04X%02X",l,at, 
				(at+l<en) ? 0 : 1 );
	
			crc = l;
			crc += ((at>>8) & 0xff);
			crc += ( at & 0xff);
			crc += (at+l<en) ? 0 : 1;
	
			for(i=0;i<l;i++) {
				fprintf(fout,"%02X",d[i]);
				crc += d[i];
			}
			crc = ~crc+1;
			fprintf(fout,"%02X\n",crc);
		} else {
			/* normal binary out */
			fwrite(d,1,l,fout);
		}

		if (verbose>1) {
			printf(".");
			fflush(stdout);
		}
	}

	close(fd);

	if (verbose)
		printf("; Done\n");

	fclose(fout);
	return 0;
}
	
